"""Constants used in SMAC, e.g. maximum number of cutoffs, very small number, etc."""
__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


MAXINT = 2**31 - 1
MINIMAL_COST_FOR_LOG = 0.00001
MAX_CUTOFF = 65535
VERY_SMALL_NUMBER = 1e-10  # The very small number is used to avoid numerical instabilities or dividing by zero
N_TREES = 10
